from unittest import TestCase

import testdata

from stockton.dns import Domain


class DomainTest(TestCase):
    def test_mx(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} mail is handled by 10 mailstore1.secureserver.net.".format(d.host),
                "{} mail is handled by 0 smtp.secureserver.net.".format(d.host),
                "",
            ])
        testdata.patch(d, query=monkey_query)

        mxs = d.mx()
        self.assertEqual(2, len(mxs))

        mxs = d.mx("smtp.secureserver.net")
        self.assertEqual(1, len(mxs))

        vs = d.nameservers("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_nameservers(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} name server ns62.domaincontrol.com.".format(d.host),
                "{} name server ns61.domaincontrol.com.".format(d.host),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.nameservers()
        self.assertEqual(2, len(vs))

        vs = d.nameservers("ns62.domaincontrol.com")
        self.assertEqual(1, len(vs))

        vs = d.nameservers("000.00.000.000")
        self.assertEqual(0, len(vs))


    def test_a(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} has address 100.70.100.200".format(d.host),
                "{} has address 101.71.101.201".format(d.host),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.a()
        self.assertEqual(2, len(vs))

        vs = d.a("101.71.101.201")
        self.assertEqual(1, len(vs))

        vs = d.a("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_txt(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} descriptive text \"google-site-verification=dakfkdkaureurjkfdhiqerukSHA\"".format(d.host),
                "{} descriptive text \"v=spf1 include:_spf.google.com include:sendgrid.net include:spf.mandrillapp.com ~all\"".format(d.host),
                "{} descriptive text \"v=spf1 include:servers.mcsv.net ?all\"".format(d.host),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.txt()
        self.assertEqual(3, len(vs))

        vs = d.txt("spf")
        self.assertEqual(2, len(vs))

        vs = d.txt("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_ptr(self):
        d = Domain("123.456.789.100")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "203.68.131.104.in-addr.arpa domain name pointer {}.".format("example.com"),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.ptr()
        self.assertEqual(1, len(vs))

        vs = d.ptr("example.com")
        self.assertEqual(1, len(vs))

        vs = d.ptr("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_rdns(self):
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "123.456.789.100.in-addr.arpa domain name pointer {}.".format(d.host),
                ""
            ])
        MonkeyDomain = testdata.patch(
            Domain,
            a=lambda *a, **kw: ["123.456.789.100"],
            query=monkey_query
        )

        d = MonkeyDomain("example.com")
        vs = d.rdns()
        self.assertEqual("123.456.789.100", vs[0])


# class MailserverTest(TestCase):
#     def test_




import re

from captain import echo

from .. import cli
from ..path import Filepath, Dirpath
from ..concur.formats.postfix import Main, SMTPd, Master
from ..concur.formats.generic import SpaceConfig

class Postfix(object):

    @property
    def mailserver(self):
        m = self.main()
        return m["myhostname"].val

    @property
    def main_f(self):
        return Filepath(Main.dest_path)

    @property
    def master_f(self):
        return Filepath(Master.dest_path)

    @property
    def config_d(self):
        return Dirpath("/etc/postfix")

    @property
    def virtual_d(self):
        return Dirpath(self.config_d, "virtual")

    @property
    def addresses_d(self):
        return Dirpath(self.virtual_d, "addresses")

    @property
    def domains_f(self):
        return Filepath(self.virtual_d, "domains")

    @property
    def domains(self):
        """return all the domains postfix has configured"""
        try:
            domains = set(self.domains_f.lines())
        except IOError:
            domains = set()
        return domains

    @property
    def addresses(self):
        for domain in self.domains:
            yield self.address(domain)

    def add_domain(self, domain, proxy_file, proxy_email):
        if not proxy_file:
            if not domain or not proxy_email:
                raise ValueError("Either pass in proxy_file or (domain and proxy_emails)")

        # create directory /etc/postfix/virtual
        virtual_d = self.virtual_d
        virtual_d.create()

        # create addresses directory
        addresses_d = self.addresses_d
        addresses_d.create()

        domains_f = self.domains_f
        domains_f.create()
        old_domains = self.domains
        new_domains = set()

        if proxy_file:
            echo.h3("Adding domain {} with addresses from {} to postfix", domain, proxy_file)
            f = Filepath(proxy_file)
            domain_f = Filepath(addresses_d, domain)
            f.copy(domain_f)
            new_domains.add(domain)

        elif proxy_email:
            echo.h3("Adding catchall for {} routing to {}", domain, proxy_email)
            domain_f = Filepath(addresses_d, domain)
            domain_c = SpaceConfig(dest_path=domain_f.path)
            domain_c["@{}".format(domain)] = proxy_email
            domain_c.save()
            new_domains.add(domain)

        domains_f.writelines(old_domains.union(new_domains))

        # this is an additive editing of the conf file, so we use the active conf
        # as the prototype
        m = self.main()
        m.update(
            ("virtual_alias_domains", domains_f.path),
            ("virtual_alias_maps", ",\n  ".join(("hash:{}".format(af) for af in self.addresses)))
        )
        m.save()

        for domain in new_domains:
            cli.run("postmap {}".format(self.address(domain)))

    def address(self, domain):
        addresses_f = Filepath(self.addresses_d, domain)
        return addresses_f

    def main(self, *path):
        """return the main config

        *path -- string -- pass in "" to not load any prototype, pass in a path
            to load that path as the prototype, pass in nothing to get the live
            main.cf file as the prototype

        return -- Main
        """
        if path:
            if path[0]:
                m = Main(prototype_path=path[0])

            else:
                m = Main()

        else:
            m = Main(prototype_path=Main.dest_path)

        return m

    def restart(self):
        try:
            cli.run("postfix status")

        except RuntimeError:
            cli.run("postfix start")

        finally:
            cli.run("postfix reload")

    def reset(self, **kwargs):
        delete_files = kwargs.get("really_delete_files", False)
        if not delete_files:
            raise ValueError("You want to delete files? pass in really_delete_files=True")

        # remove any .bak files in this directory, we do this to make sure any other
        # commands that configure postfix will be able to make correct snapshots of
        # the main.cf file to remain idempotent
        echo.h3("Clearing .bak files")
        postfix_d = Dirpath("/etc/postfix")
        postfix_d.delete_files(".bak$")

        virtual_d = self.virtual_d
        echo.h3("Clearing {}", virtual_d)
        virtual_d.clear()





import re

from captain import echo

from .. import cli
from .postfix import Postfix
from ..path import Filepath, Dirpath
from ..concur.formats.opendkim import OpenDKIM


class DomainKey(object):
    @property
    def text(self):
        dkim_text = "{} {} {}".format(self.v, self.k, self.p)
        return dkim_text

    def __init__(self, domain):
        self.domain = domain

        dk = DKIM()
        txt_f = Filepath(dk.keys_d, "{}.txt".format(domain))
        contents = txt_f.contents()
        m = re.match("^(\S+)", contents)
        self.subdomain = "{}.{}".format(m.group(1), domain)

        mv = re.search("v=\S+", contents)
        mk = re.search("k=\S+", contents)
        mp = re.search("p=[^\"]+", contents)
        self.v = mv.group(0)
        self.k = mk.group(0)
        self.p = mp.group(0)

    def __str__(self):
        return self.text


class DKIM(object):

    @property
    def config_f(self):
        return Filepath(OpenDKIM.dest_path)

    @property
    def opendkim_d(self):
        return Dirpath("/etc/opendkim")

    @property
    def keys_d(self):
        return Dirpath(self.opendkim_d, "keys")

    @property
    def keytable_f(self):
        return Filepath(self.opendkim_d, "KeyTable")

    @property
    def signingtable_f(self):
        return Filepath(self.opendkim_d, "SigningTable")

    @property
    def trustedhosts_f(self):
        return Filepath(self.opendkim_d, "TrustedHosts")

    def domainkey(self, domain):
        return DomainKey(domain)

    def add_domains(self):
        #     opendkim_d = Dirpath("/etc/opendkim")
        #     keys_d = Dirpath(opendkim_d, "keys")
        #     keytable_f = Filepath(opendkim_d, "KeyTable")
        #     keytable_f.clear()
        #     signingtable_f = Filepath(opendkim_d, "SigningTable")
        #     signingtable_f.clear()
        #     trustedhosts_f = Filepath(opendkim_d, "TrustedHosts")
        #     trustedhosts_f.clear()

        p = Postfix()
        for domain in p.domains:
            self.add_domain(domain)

    def add_domain(self, domain, gen_key=False):
        echo.h3("Configuring DKIM for {}", domain)

        opendkim_d = self.opendkim_d
        keys_d = self.keys_d
        keytable_f = self.keytable_f
        signingtable_f = self.signingtable_f
        trustedhosts_f = self.trustedhosts_f

        private_f = Filepath(keys_d, "{}.private".format(domain))
        txt_f = Filepath(keys_d, "{}.txt".format(domain))
        if not txt_f.exists() or gen_key:
            #cli.run("opendkim-genkey --domain={} --verbose --directory=\"{}\"".format(
            cli.run("opendkim-genkey --bits=2048 --domain={} --directory=\"{}\"".format(
                domain,
                keys_d.path
            ))

            private_f = Filepath(keys_d, "default.private")
            private_f.rename("{}.private".format(domain))
            private_f.chmod(600)
            private_f.chown("opendkim:opendkim")

            txt_f = Filepath(keys_d, "default.txt")
            txt_f.rename("{}.txt".format(domain))

            dk = self.domainkey(domain)

        if not keytable_f.contains(domain):
            keytable_f.append("{} {}:default:{}\n".format(
                dk.subdomain,
                domain,
                private_f.path
            ))

        if not signingtable_f.contains(domain):
            signingtable_f.append("{} {}\n".format(
                domain,
                dk.subdomain
            ))

        if not trustedhosts_f.contains(domain):
            trustedhosts_f.append("*.{}\n".format(domain))


    def restart(self):
        output = cli.run("/etc/init.d/opendkim status", capture_output=True)
        if re.search("opendkim\s+is\s+running", output, flags=re.I):
            cli.run("/etc/init.d/opendkim start")

        else:
            cli.run("/etc/init.d/opendkim restart")





"""
The base classes for all the other formats,

usually, you would actually use the classes in generics.py as the start for your custom
configuration parser, the classes in this module are the support classes for the
generics
"""
import re
from collections import defaultdict


class ConfigBase(object):
    """the base common class for most of the other more useful classes"""
    def __init__(self, config):
        self.config = config
        self.reset()

    def reset(self):
        self.sections = defaultdict(list)
        self.options = defaultdict(list)
        self.lines = []
        self.name = ""
        self.val = ""

    def parse(self, fp):
        self.line = fp.line
        self._parse(fp)
        self.modified = False

    def is_valid(self):
        return True

    def update(self, *args, **kwargs):
        for k, v in args:
            self[k] = v

        for k, v in kwargs.items():
            self[k] = v

    def __setattr__(self, k, v):
        # http://stackoverflow.com/questions/17576009/python-class-property-use-setter-but-evade-getter
        if k == "modified":
            self.__dict__["modified"] = v

        else:
            self.modified = True
            super(ConfigBase, self).__setattr__(k, v)

    def __missing__(self, k):
        try:
            # we are either a section or the config option, assume we are a
            # section, fallback to being the config
            option = self.config.create_option()
        except AttributeError:
            option = self.create_option()

        option.name = k
        self.options[option.name].append(len(self.lines))
        self.lines.append(option)
        return option

    def __contains__(self, k):
        return k in self.sections or k in self.options

    def __getitem__(self, k):
        line_numbers = []
        if k in self.sections:
            line_numbers = self.sections[k]

        elif k in self.options:
            line_numbers = self.options[k]

        if len(line_numbers) == 1:
            v = self.lines[line_numbers[0]]

        elif len(line_numbers) > 1:
            v = [self.lines[ln] for ln in line_numbers]

        else:
            v = self.__missing__(k)
            #raise KeyError("no section or option {}".format(k))

        return v

    def __setitem__(self, k, v):
        if k in self.sections:
            raise ValueError("you cannot modify sections using dict notation")

        line_numbers = []

        if k in self.options:
            line_numbers = self.options[k]

        if len(line_numbers) > 0:
            for line_number in line_numbers:
                option = self.lines[line_number]
                option.val = v

        else:
            option = self.__missing__(k)
            option.val = v


class ConfigLine(ConfigBase):
    def __str__(self):
        return self.line

    def _parse(self, fp):
        self.line = fp.line


class ConfigOption(ConfigLine):
    def is_valid(self):
        return bool(self.name)

    def _parse(self, fp):
        line = fp.line
        name = ""
        val = ""
        comment = ""
        commenters = self.config.commenters
        divider = self.config.option_divider

        if re.match("^[^{}]\S+".format(commenters), line): # name = val
            name, val = re.split("\s*{}\s*".format(divider), line, 2)

        elif re.match("^[{}]\s*\S+\s*{}".format(commenters, divider), line): # # name = val
            l = re.sub("^[{}]\s*".format(commenters), "", line)
            name, val = re.split("\s*{}\s*".format(divider), l, 2)

        bits = re.split("\s[{}]\s*".format(commenters), val, 2)
        val = bits[0]
        if len(bits) > 1:
            comment = bits[1]

        self.name = name
        self.val = val
        self.comment = comment
        self.line = line
        self.modified = False

    def __str__(self):
        s = ""
        if self.modified:
            s = self.format_set(
                self.name,
                self.config.option_divider,
                self.val
            )
            comment = getattr(self, "comment", "")
            if comment:
                s += " {} {}".format(self.commenters[0], comment)

        else:
            s = super(ConfigOption, self).__str__()

        return s

    def format_set(self, name, divider, val):
        option_set_format = "{} {} {}" # key, divider, val
        s = option_set_format.format(name, divider, val)
        return s


class ConfigSection(ConfigOption):
    def _parse(self, fp):
        pass

    def __str__(self):
        s = "\n".join(str(cl) for cl in self.lines)
        return s


class ConfigFile(object):
    def __init__(self, path, config):
        self.path = path
        self.config = config
        # we load the whole config file into memory because we hate memory and to
        # punish it we like to use lots of it. Also, we were having some difficulty
        # getting some conf files to parse correctly without being able to seek and
        # stuff to certain lines and replay parts of the file and this was the 
        # easiest solution to that problem
        with open(self.path, "r") as fp:
            self.lines = fp.readlines()
            self.count = len(self.lines)

        self.line_number = -1

    def close(self):
        if not self.fp.closed:
            self.fp.close()

    def __iter__(self):
        return self

    def __count__(self):
        return self.count

    def rewind(self, line_number):
        self.line_number = line_number

    def next(self):
        self.line_number += 1
        if self.line_number >= self.count:
            raise StopIteration()

        self.line = self.lines[self.line_number].rstrip()

        c = self.config.create_section()
        c.parse(self)
        if not c.is_valid():
            c = self.config.create_option()
            c.parse(self)
            if not c.is_valid():
                c = self.config.create_line()
                c.parse(self)

        return c


class Config(ConfigBase):
    """config parser that allows you to modify the parsed file and write it back out

    we don't use shlex because it didn't preserve comments
    https://docs.python.org/2/library/shlex.html 

    If an ini specific parser is added at a later date, use this?
    https://docs.python.org/2/library/configparser.html
    """

    commenters = "#"

    option_divider = "="

    line_class = ConfigLine

    option_class = ConfigOption

    section_class = ConfigSection

    file_class = ConfigFile


    dest_path = ""

    prototype_path = ""

    def __init__(self, dest_path="", prototype_path=""):
        self.reset()

        if prototype_path:
            self.prototype_path = prototype_path
            self.parse()

        if dest_path:
            self.dest_path = dest_path

    def create_line(self):
        return self.line_class(self)

    def create_option(self):
        return self.option_class(self)

    def create_section(self):
        return self.section_class(self)

    def create_file(self):
        return self.file_class(self.prototype_path, self)

    def parse(self):
        fp = self.create_file()
        self.reset()

        for line_number, c in enumerate(fp):
            if isinstance(c, self.section_class):
                self.sections[c.name].append(line_number)
                self.lines.append(c)

            elif isinstance(c, self.option_class):
                self.options[c.name].append(line_number)
                self.lines.append(c)
 
            elif isinstance(c, self.line_class):
                self.lines.append(c)

    def __str__(self):
        return "\n".join((str(cl) for cl in self))

    def __iter__(self):
        for cl in self.lines:
            yield cl

    def save(self):
        with open(self.dest_path, "w") as f:
            for cl in self:
                f.write(str(cl))
                f.write("\n")




from unittest import TestCase
import os
import re

import testdata
from captain.client import Captain

from stockton import cli
from stockton.path import Filepath, Dirpath
from stockton.concur.formats.postfix import Main, SMTPd, Master


class Stockton(Captain):
    def __init__(self, subcommand):
        self.cmd_prefix = "python -m stockton --verbose {}".format(subcommand.replace("_", "-"))
        super(Stockton, self).__init__("")

    def run(self, arg_str='', **process_kwargs):
        lines = ""
        for line in super(Stockton, self).run(arg_str, **process_kwargs):
            self.flush(line)
            lines += line
        return lines


def setUpModule():
    if os.environ["USER"] != "root":
        raise RuntimeError("User is not root, re-run this test with sudo")


class PrepareTest(TestCase):
    def setUp(self):
        cli.purge("postfix")

    def test_run(self):
        d = Dirpath("etc", "postfix")
        self.assertFalse(d.exists())

        s = Stockton("prepare")
        r = s.run("")

        # we use a file to get around filecache at os level
        f = Filepath(Main.dest_path)
        self.assertTrue(f.exists())


class ConfigureTest(TestCase):
    def setUp(self):
        f = Filepath(Main.dest_path)
        self.assertTrue(f.exists())

    def test_recv(self):
        s = Stockton("configure-recv")

        with self.assertRaises(RuntimeError):
            r = s.run("example.com --mailserver=mail.example.com")

        arg_str = "example.com --mailserver=mail.example.com --proxy-email=final@destination.com"
        r = s.run(arg_str)

        # make some changes to main
        m = Main(prototype_path=Main.dest_path)
        m.update(
            ("foobar", "1234")
        )
        m.save()

        f = Filepath(Main.dest_path)
        self.assertTrue(f.contains("foobar"))

        # re-run
        s = Stockton("configure-recv")
        r = s.run(arg_str)
        f = Filepath(Main.dest_path)
        self.assertFalse(f.contains("foobar"))

        cli.running("postfix")

    def test_send(self):

        cli.purge("sasl2-bin", "libsasl2-modules")

        cli.package("cyrus-clients-2.4") # for smtptest

        s = Stockton("configure-send")
        arg_str = "example.com --mailserver=mail.example.com --smtp-password=1234 --state=CA --city=\"San Francisco\""
        r = s.run(arg_str)

        r = cli.run(
            "echo QUIT | smtptest -a smtp@example.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
            capture_output=True
        )
        self.assertRegexpMatches(r, "235[^A]+Authentication\s+successful")

        r = cli.run(
            "echo QUIT | smtptest -a smtp@example.com -w 9876 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
            capture_output=True
        )
        self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")

        r = cli.run(
            "echo QUIT | smtptest -a foo@example.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
            capture_output=True
        )
        self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")

#         r = cli.run(
#             "echo QUIT | smtptest -a smtp -w 1234 -t /etc/postfix/certs/example.com.pem -p 587 localhost",
#             capture_output=True
#         )
#         self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")

        cli.running("postfix")

    def test_dkim(self):
        #self.test_recv() # we need a configured for receive postfix
        s = Stockton("configure-recv")
        r = s.run("example.com --mailserver=mail.example.com --proxy-email=final@destination.com")


        cli.purge("opendkim", "opendkim-tools")
        opendkim_d = Dirpath("/etc/opendkim")
        opendkim_d.delete()

        s = Stockton("configure-dkim")
        r = s.run()

        self.assertTrue(opendkim_d.exists())

        keytable_f = Filepath(opendkim_d, "KeyTable")
        self.assertEqual(1, keytable_f.lc())

        signingtable_f = Filepath(opendkim_d, "SigningTable")
        self.assertEqual(1, signingtable_f.lc())

        trustedhosts_f = Filepath(opendkim_d, "TrustedHosts")
        self.assertTrue("*.example.com" in trustedhosts_f.lines())

        cli.running("postfix")
        cli.running("opendkim")

    def test_srs(self):
        s = Stockton("configure-srs")
        r = s.run("")

        cli.running("postsrsd")


class DomainTest(TestCase):
    def setUp(self):
        f = Filepath(Main.dest_path)
        self.assertTrue(f.exists())

    def test_delete_domain(self):
        s = Stockton("configure-recv")
        r = s.run("example.com --mailserver=mail.example.com --proxy-email=final@dest.com")
        s = Stockton("configure_dkim")
        r = s.run()

        s = Stockton("add-domain")
        r = s.run("todelete.com --proxy-email=todelete@dest.com")

        s = Stockton("delete-domain")
        r = s.run("todelete.com")

        # verify
        af = Filepath("/etc/postfix/virtual/addresses/todelete.com")
        self.assertFalse(af.exists())

        df = Filepath("/etc/postfix/virtual/domains")
        self.assertFalse(df.contains("todelete.com"))

        opendkim_d = Dirpath("/etc/opendkim")
        #keytable_f = Filepath(opendkim_d, "KeyTable")
        signingtable_f = Filepath(opendkim_d, "SigningTable")
        self.assertFalse(signingtable_f.contains("todelete.com"))

        trustedhosts_f = Filepath(opendkim_d, "TrustedHosts")
        self.assertFalse(trustedhosts_f.contains("todelete.com"))

    def test_add_domain_proxy_file_smtp(self):
        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")
        s = Stockton("configure-send")
        arg_str = "example.com --mailserver=mail.example.com --smtp-password=1234 --state=CA --city=\"San Francisco\""
        r = s.run(arg_str)

        proxy_addresses = testdata.create_file("foo.com", [
            "one@foo.com                foo@dest1.com",
            "two@foo.com                foo@dest2.com",
            "three@foo.com              foo@dest3.com",
            "",
        ])

        s = Stockton("add-domain")
        r = s.run("foo.com --proxy-file={} --smtp-password=12345".format(proxy_addresses))

        cli.package("cyrus-clients-2.4") # for smtptest
        r = cli.run(
            "echo QUIT | smtptest -a smtp@foo.com -w 12345 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
            capture_output=True
        )
        self.assertRegexpMatches(r, "235[^A]+Authentication\s+successful")

    def test_add_domain_proxy_file(self):
        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")

        proxy_domains = testdata.create_files({
            "foo.com": "\n".join([
                "one@foo.com                foo@dest.com",
                "two@foo.com                foo@dest.com",
                "three@foo.com              foo@dest.com",
                "",
            ]),
            "bar.com": "\n".join([
                "one@bar.com                bar@dest.com",
                "two@bar.com                bar@dest.com",
                "three@bar.com              bar@dest.com",
                "",
            ]),
        })

        for f in proxy_domains:
            s = Stockton("add-domain")
            r = s.run("{} --proxy-file={}".format(f.name, f))

        contents = "\n".join([
            "@foo.com                   foo@dest.com",
            "",
        ])
        f = testdata.create_file("foo.com", contents)

        r = s.run("{} --proxy-file={}".format(f.name, f))
        foo = Filepath("/etc/postfix/virtual/addresses/foo.com")
        self.assertEqual(contents, foo.contents())

        # Let's check structure because I was having a lot of problems with getting
        # the structure...

        # domains file should have 2 domains in it
        domains_f = Filepath("/etc/postfix/virtual/domains")
        self.assertEqual(2, domains_f.lc())
        self.assertTrue(domains_f.contains("^foo\.com$"))

        m = Main(prototype_path=Main.dest_path)
        # we can't guarrantee foo, bar order so we match one line at a time
        self.assertTrue(re.search("^hash:[/a-z]+?(bar|foo)\.com,", m["virtual_alias_maps"].val, re.M))
        self.assertTrue(re.search("^\s+hash:[/a-z]+?(bar|foo)\.com$", m["virtual_alias_maps"].val, re.M))

    def test_add_domain_domain(self):
        """pass in the domain and the proxy"""
        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")
        s = Stockton("configure-dkim")
        r = s.run()

        s = Stockton("add-domain")
        onef = Filepath("/etc/postfix/virtual/addresses/one.com")
        twof = Filepath("/etc/postfix/virtual/addresses/two.com")
        df = Filepath("/etc/postfix/virtual/domains")

        r = s.run("one.com --proxy-email=one@dest.com")
        self.assertTrue(onef.contains("^@one.com"))
        self.assertTrue(df.contains("one.com"))

        r = s.run("two.com --proxy-email=two@dest.com")
        self.assertTrue(onef.contains("^@one.com"))
        self.assertTrue(df.contains("one.com"))
        self.assertTrue(twof.contains("two@dest.com"))
        self.assertTrue(df.contains("two.com"))

        opendkim_d = Dirpath("/etc/opendkim")
        keytable_f = Filepath(opendkim_d, "KeyTable")
        signingtable_f = Filepath(opendkim_d, "SigningTable")
        trustedhosts_f = Filepath(opendkim_d, "TrustedHosts")

        self.assertEqual(2, keytable_f.lc())
        self.assertEqual(2, signingtable_f.lc())
        #self.assertTrue("*.example.com" in trustedhosts_f.lines())
        self.assertTrue("*.one.com" in trustedhosts_f.lines())
        self.assertTrue("*.two.com" in trustedhosts_f.lines())

    def test_add_domain_smtp(self):
        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")
        s = Stockton("configure-send")
        arg_str = "example.com --mailserver=mail.example.com --smtp-password=1234 --state=CA --city=\"San Francisco\""
        r = s.run(arg_str)

        s = Stockton("add-domain")
        r = s.run("one.com --proxy-email=one@dest.com --smtp-password=1234")

        cli.package("cyrus-clients-2.4") # for smtptest
        r = cli.run(
            "echo QUIT | smtptest -a smtp@one.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
            capture_output=True
        )
        self.assertRegexpMatches(r, "235[^A]+Authentication\s+successful")


class LockdownTest(TestCase):
    def setUp(self):
        cli.running("postfix")

    def test_run(self):
        s = Stockton("configure-recv")
        r = s.run("example.com --mailserver=mail.example.com --proxy-email=final@dest.com")

        s = Stockton("lockdown")
        r = s.run("--mailserver=mail.example.com")
        cli.running("postfix")

        helo_f = Filepath("/etc/postfix/helo.regexp")
        self.assertTrue(helo_f.exists())




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import argparse
import os
import subprocess
import re

from captain import echo, exit, ArgError
from captain.decorators import arg, args

from stockton import __version__, cli, dns
from stockton.concur.formats.postfix import Main, SMTPd, Master
from stockton.concur.formats.opendkim import OpenDKIM
from stockton.concur.formats.generic import SpaceConfig
from stockton.path import Dirpath, Filepath, Sentinal

from stockton.interface import SMTP, Postfix, DKIM


def main_prepare():
    echo.h2("Installing Postfix")

    with Sentinal.check("apt-get-update") as s:
        if not s:
            cli.run("apt-get update")

    #cli.run("apt-get -y install --no-install-recommends postfix")
    #cli.run("debconf-set-selections <<< \"postfix postfix/main_mailer_type string 'No configuration'\"")

    cli.package("postfix")


@arg('domain', default="", help='The email domain (eg, example.com)')
@arg('--mailserver', help='The domain mailserver (eg, mail.example.com)')
@arg('--proxy-file', default="", help='The file containing domain addresses to proxy emails')
@arg('--proxy-email', default="", help='The final destination email address')
def main_configure_recv(domain, mailserver, proxy_file, proxy_email):
    # http://www.postfix.org/VIRTUAL_README.html

    echo.h2("Configuring Postfix to receive emails")

    p = Postfix()
    p.reset(really_delete_files=True)

    m = p.main("")
    m.update(
        ("alias_maps", "hash:/etc/aliases"), # http://unix.stackexchange.com/a/244200/118750
        ("myhostname", mailserver),
        #("mydomain", domain),
        #("myorigin", domain),
    )
    m.save()

    if domain:
        p.add_domain(domain, proxy_file, proxy_email)

    # make backup of master.cf
    master_bak = p.master_f.backup(ignore_existing=False)

    p.restart()


@arg('domain', help='The email domain (eg, example.com)')
@arg('--mailserver', help='The domain mailserver (eg, mail.example.com)')
@arg('--smtp-username', default="smtp", help='smtp username for sending emails')
@arg('--smtp-password', help='smtp password for sending emails')
@arg('--country', default="US", help='country for ssl certificate')
@arg('--state', help='state for ssl certificate')
@arg('--city', help='city for ssl certificate')
def main_configure_send(domain, mailserver, smtp_username, smtp_password, country, state, city):

    # https://help.ubuntu.com/lts/serverguide/postfix.html#postfix-sasl
    # http://www.postfix.org/SASL_README.html

    echo.h2("Configuring Postfix to send emails")

    cli.package("sasl2-bin", "libsasl2-modules")

    s = SMTP()
    s.add_user(smtp_username, smtp_password, domain)

    s = SMTPd()
    s.update(
        ("pwcheck_method", "auxprop"),
        ("auxprop_plugin", "sasldb"),
        ("mech_list", "PLAIN LOGIN CRAM-MD5 DIGEST-MD5 NTLM"),
        ("log_level", "7")
    )
    s.save()

    certs_d = Dirpath("/etc/postfix/certs")
    certs_d.create()
    certs_key = Filepath(certs_d, "{}.key".format(mailserver))
    certs_crt = Filepath(certs_d, "{}.crt".format(mailserver))
    certs_pem = Filepath(certs_d, "{}.pem".format(mailserver))

    cli.package("openssl", only_upgrade=True)

    # http://superuser.com/questions/226192/openssl-without-prompt
    cli.run(" ".join([
        "openssl req",
        "-new",
        "-newkey rsa:4096",
        # openssl 1.0.2+ only, comment out above line and uncomment next 2
        #"-newkey ec",
        #"-pkeyopt ec_paramgen_curve:prime256v1",
        "-days 3650",
        "-nodes",
        "-x509",
        "-subj \"/C={}/ST={}/L={}/O={}/CN={}\"".format(
            country,
            state,
            city,
            mailserver,
            mailserver
        ),
        "-keyout {}".format(certs_key),
        "-out {}".format(certs_crt)
    ]))
    cli.run("cat {} {} > {}".format(certs_crt, certs_key, certs_pem))

    # make backup of master.cf if it doesn't already exist
    master_f = Filepath(Master.dest_path)
    master_bak = master_f.backup(ignore_existing=False)

    # add the config to master.cf to enable smtp sending
    m = Master(prototype_path=master_bak.path)
    for smtp in m["smtp"]:
        if smtp.cmd == "smtpd":
            smtp.chroot = "n"

    m["submission"].chroot = "n"
    m["submission"].update(
        ("syslog_name", "postfix/submission"),
        ("smtpd_tls_security_level", "may"),
        ("smtpd_tls_cert_file", certs_pem.path),
        ("smtpd_sasl_auth_enable", "yes"),
        ("smtpd_reject_unlisted_recipient", "no"),
        ("smtpd_relay_restrictions", "permit_sasl_authenticated,reject"),
        ("milter_macro_daemon_name", "ORIGINATING")
    )
    m.save()

    cli.postfix_reload()


def main_configure_dkim():
    """DomainKeys Identified Mail (DKIM)

    Setup dkim support

    http://www.dkim.org/
    """

    # setup for multi-domain support thanks to
    # http://askubuntu.com/questions/438756/using-dkim-in-my-server-for-multiple-domains-websites
    # http://seasonofcode.com/posts/setting-up-dkim-and-srs-in-postfix.html
    # http://edoceo.com/howto/opendkim
    # https://help.ubuntu.com/community/Postfix/DKIM

    echo.h2("Configuring Postfix to use DKIM")

    cli.package("opendkim", "opendkim-tools")

    dk = DKIM()
    opendkim_d = dk.opendkim_d
    opendkim_d.create()
    opendkim_d.clear()

    keys_d = dk.keys_d
    keys_d.create()
    keys_d.clear()

    hosts = [
        "127.0.0.1",
        "::1",
        "localhost",
        "192.168.0.1/24",
    ]
    # could also use https://pypi.python.org/pypi/netifaces but this seemed easier
    # http://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
    external_ip = cli.ip()
    if external_ip:
        hosts.append(external_ip)

    hosts.append("")
    trustedhosts_f = dk.trustedhosts_f
    trustedhosts_f.writelines(hosts)
    #trustedhosts_f = opendkim_d.create_file("TrustedHosts", "\n".join(hosts))

    keytable_f = dk.keytable_f
    signingtable_f = dk.signingtable_f

    # make backup of config
    config_f = dk.config_f
    config_bak = config_f.backup(ignore_existing=False)

    c = OpenDKIM(prototype_path=config_bak.path)
    c.update(
        ("Canonicalization", "relaxed/simple"),
        ("Mode", "sv"),
        ("SubDomains", "yes"),
        ("Syslog", "yes"),
        ("LogWhy", "yes"),
        ("UMask", "022"),
        ("UserID", "opendkim:opendkim"),
        ("KeyTable", keytable_f.path),
        ("SigningTable", signingtable_f.path),
        ("ExternalIgnoreList", trustedhosts_f.path),
        ("InternalHosts", trustedhosts_f.path),
        ("Socket", "inet:8891@localhost")
    )
    c.save()

    p = Postfix()
    m = p.main()
    m.update(
        ("milter_default_action", "accept"),
        # http://lists.opendkim.org/archive/opendkim/users/2011/08/1297.html
        ("milter_protocol", "6"),
        ("smtpd_milters", "inet:localhost:8891"),
        ("non_smtpd_milters", "inet:localhost:8891"),
    )
    m.save()

    # http://unix.stackexchange.com/a/74491/118750
    cli.run("adduser postfix opendkim")

    dk.add_domains()

    p.restart()
    dk.restart()


def main_configure_srs():
    """SRS: Sender Rewriting Scheme

    This configures SRS so SPF will work correctly and treat this MTA as a proxy

    http://www.openspf.org/SRS
    """

    # thanks to
    # http://seasonofcode.com/posts/setting-up-dkim-and-srs-in-postfix.html
    # http://serverfault.com/questions/82234/srs-sender-rewriting-when-forwarding-mail-through-postfix
    # http://www.mind-it.info/2014/02/22/forward-postfix-spf-srs/
    # http://www.openspf.org/SRS

    echo.h2("Configuring Postfix to use SRS")

    cli.package("unzip", "cmake", "curl", "build-essential")

    tmpdir = Dirpath.get_temp()

    # only download the srs if it doesn't already exist
    postsrs_f = Filepath(tmpdir, "postsrsd.zip")
    with Sentinal.check("srs") as s:
        if not s:
            # https://github.com/roehling/postsrsd
            cli.run("curl -L -o postsrsd.zip https://github.com/roehling/postsrsd/archive/master.zip", cwd=tmpdir.path)

        else:
            echo.out("****** not downloading postsrsd because sentinal {}", s)

    # Build and install.
    cli.run("unzip -o postsrsd.zip", cwd=tmpdir.path)
    build_d = Dirpath(tmpdir, "postsrsd-master", "build")
    build_d.create()

    cli.run("cmake -DCMAKE_INSTALL_PREFIX=/usr ../", cwd=build_d.path)
    cli.run("make", cwd=build_d.path)
    cli.run("make install", cwd=build_d.path)

    m = Main(prototype_path=Main.dest_path)
    m.update(
        ("sender_canonical_maps", "tcp:localhost:10001"),
        ("sender_canonical_classes", "envelope_sender"),
        ("recipient_canonical_maps", "tcp:localhost:10002"),
        ("recipient_canonical_classes", "envelope_recipient,header_recipient")
    )
    m.save()

    cli.srs_reload()
    cli.postfix_reload()


@arg('domain', help='The email domain to remove (eg, example.com)')
def main_delete_domain(domain):
    """remove a domain from the server"""
    echo.h3("Deleting domain {}", domain)

    # TODO -- most of this should be moved to postfix.py and dkim.py files

    m = Main(prototype_path=Main.dest_path)

    # remove postfix settings
    virtual_d = Dirpath("/etc/postfix/virtual")

    domains_f = Filepath(virtual_d, "domains")
    domains_f.delete_lines("^{}$".format(domain))

    addresses_d = Dirpath(virtual_d, "addresses")
    addresses_d.delete_files("^{}".format(domain))

    domain_hashes = []
    for d in domains_f.lines():
        domain_f = Filepath(addresses_d, d)
        domain_hashes.append("hash:{}".format(domain_f.path))

    m = Main(prototype_path=Main.dest_path)
    m["virtual_alias_maps"] = ",\n  ".join(domain_hashes)
    m.save()

    # make sure lockdown is still idempotent
    main_f = Filepath(Main.dest_path)
    main_bak = main_f.backup(suffix=".lockdown.bak", ignore_existing=True)

    # remove dkim settings
    opendkim_d = Dirpath("/etc/opendkim")

    signingtable_f = Filepath(opendkim_d, "SigningTable")
    signingtable_f.delete_lines(domain)

    trustedhosts_f = Filepath(opendkim_d, "TrustedHosts")
    trustedhosts_f.delete_lines(domain)

    cli.postfix_reload()
    cli.opendkim_reload()


@arg('domain', default="", help='The email domain (eg, example.com)')
@arg('--proxy-file', default="", help='The file containing domain addresses to proxy emails')
#@arg('--proxy-domains', default="", help='The directory containing domain configuration files')
@arg('--proxy-email', default="", help='The final destination email address')
@arg('--smtp-username', default="smtp", help='smtp username for sending emails')
@arg('--smtp-password', default="", help='smtp password for sending emails')
def main_add_domain(domain, proxy_file, proxy_email, smtp_username, smtp_password):

    p = Postfix()
    p.add_domain(domain, proxy_file, proxy_email)

    dk = DKIM()
    dk.add_domain(domain)

    if smtp_password:
        s = SMTP()
        s.add_user(smtp_username, smtp_password, domain)

    p.restart()
    dk.restart()

    main_check_domain(domain)


def main_dkim_genkeys():
    dk = DKIM()
    p = Postfix()
    domains_f = Filepath("/etc/postfix/virtual/domains")
    for domain in p.domains:
        dk.add_domain(domain, gen_key=True)
        main_check_domain(domain, ["dkim"])

    p.restart()
    dk.restart()


def disabled_configure_greylist(self):
    # TODO -- not sure if want to enable this
    # https://www.debuntu.org/postfix-and-postgrey-a-proactive-approach-to-spam-filtering/
    # https://www.debuntu.org/postfix-and-postgrey-a-proactive-approach-to-spam-filtering-page-2/
    # https://github.com/schweikert/postgrey
    # http://serverfault.com/questions/701241/postgrey-whitelist-recipients-not-working
    # http://serverfault.com/questions/1817/does-smtp-greylisting-a-stop-much-spam-and-b-stop-much-legitimate-mail?rq=1
    # http://serverfault.com/questions/436327/is-greylisting-still-an-efficient-method-for-preventing-spam?rq=1
    # http://serverfault.com/questions/655924/possible-to-configure-postgrey-to-only-graylist-com-addresses
    pass


@arg('--mailserver', help='The domain mailserver (eg, mail.example.com)')
def main_lockdown(mailserver):
    # http://www.cyberciti.biz/tips/postfix-spam-filtering-with-blacklists-howto.html
    # http://www.cyberciti.biz/faq/postfix-limit-incoming-or-receiving-email-rate/
    # https://www.debuntu.org/how-to-fight-spam-with-postfix-rbl/
    # https://www.howtoforge.com/virtual_postfix_antispam

    echo.h2("Locking down Postfix")

    # make sure we've got a backup before we start messing with stuff, we won't
    # create a backup if it already exists
    main_f = Filepath(Main.dest_path)
    main_bak = main_f.backup(suffix=".lockdown.bak", ignore_existing=False)

    external_ip = cli.ip()

    echo.h3("configuring main")

    helo_f = SpaceConfig(dest_path="/etc/postfix/helo.regexp")
    helo_f.update(
        ("/^{}$/".format(re.escape(mailserver)), "550 Don't use my own hostname"),
        ("/^{}$/".format(re.escape(external_ip)), "550 Don't use my own IP address"),
        ("/^\[{}\]$/".format(re.escape(external_ip)), "550 Don't use my own IP address"),
        ("/^[0-9.]+$/", "550 Your software is not RFC 2821 compliant"),
        ("/^[0-9]+(\.[0-9]+){3}$/", "550 Your software is not RFC 2821 compliant")
    )
    helo_f.save()

    m = Main(prototype_path=main_bak.path)
    m.update(
        ("disable_vrfy_command", "yes"),
        ("smtpd_delay_reject", "yes"),
        ("smtpd_helo_required", "yes"),
        ("strict_rfc821_envelopes", "yes"),
        ("smtpd_helo_restrictions", ",\n    ".join([
            "permit_mynetworks",
            "permit_sasl_authenticated",
            "reject_non_fqdn_hostname",
            "reject_invalid_hostname",
            "regexp:/etc/postfix/helo.regexp",
            "permit"
        ])),
        ("smtpd_recipient_restrictions", ",\n    ".join([
            "permit_mynetworks",
            "permit_sasl_authenticated",
            "reject_invalid_hostname",
            "reject_non_fqdn_hostname",
            "reject_non_fqdn_sender",
            "reject_non_fqdn_recipient",
            "reject_unknown_sender_domain",
            "reject_unknown_recipient_domain",
            "reject_unauth_destination",
            "reject_unknown_reverse_client_hostname",
            "reject_rbl_client zen.spamhaus.org",
            "reject_rbl_client bl.spamcop.net",
            "reject_rbl_client b.barracudacentral.org",
            #"reject_rbl_client cbl.abuseat.org",
            #"reject_rbl_client dul.dnsbl.sorbs.net,",
            #"reject_rhsbl_sender dsn.rfc-ignorant.org",
            "permit"
        ])),
        ("smtpd_error_sleep_time", "1s"),
        ("smtpd_soft_error_limit", "10"),
        ("smtpd_hard_error_limit", "20")
    )
    m.save()

    # dbl.spamhaus.org
    # xbl.spamhaus.org
    # b.barracudacentral.org
    # http://serverfault.com/a/514830/190381

    cli.postfix_reload()

# TODO -- I have renamed main_check_domains to domain, so things need to be fixed
# the add-domain should take a domain-addresses and there should be a new command
# add-domains that take an addresses dir (what is proxy-domains


@arg('domain', help='The domain whose dns will be checked (eg, example.com)')
@arg(
    '--record', "-r",
    action="append",
    dest="records",
    help='records to check',
    default=None,
    choices=["mx", "a", "ptr", "spf", "dkim"],
)
def main_check_domain(domain, records=None):
    if not records:
        records = ["mx", "a", "ptr", "spf", "dkim"]

    pf = Postfix()
    mailserver = pf.mailserver
    external_ip = cli.ip()

    def print_record(name, domain, records):
        if records:
            echo.h1("DNS {} RECORD NEEDED FOR {}", name, domain)
            echo.br()

            echo.columns(*zip(*records))
            echo.br()

        else:
            echo.h3("{} record found", name)
            echo.br()

    if "a" in records:
        d = dns.Mailserver(mailserver, external_ip)
        print_record("A", mailserver, d.needed_a())

    if "ptr" in records:
        d = dns.Mailserver(mailserver, external_ip)
        print_record("PTR", mailserver, d.needed_ptr())

    if "mx" in records:
        d = dns.Alias(domain, mailserver)
        print_record("MX", domain, d.needed_mx())

    if "spf" in records:
        d = dns.Alias(domain, mailserver)
        print_record("SPF", domain, d.needed_spf())

    if "dkim" in records:
        try:
            d = dns.Alias(domain, mailserver)
            print_record("DKIM", domain, d.needed_dkim())
        except IOError:
            echo.h3("No local DKIM key found for {}", domain)


@args(main_configure_recv, main_configure_send)
def main_install(domain, mailserver, smtp_username, smtp_password, country, state, city, proxy_domains, proxy_email):

    main_prepare()
    main_configure_recv(domain, mailserver, proxy_domains, proxy_email)
    main_configure_send(domain, mailserver, smtp_username, smtp_password, country, state, city)
    main_configure_dkim()
    main_configure_srs()
    main_lockdown(mailserver)
    main_check_domain(domain)


def console():
    """we wrap captain.exit() so we can check for root"""
    if os.environ["USER"] != "root":
        raise RuntimeError("User is not root, re-run command with sudo")

    exit()

console()






from .. import cli
from ..path import Filepath
#from ..concur.formats.postfix import Main, SMTPd, Master


class SMTP(object):

    def add_user(self, username, password, domain):
        """adds an smtp user with credentials:

            user: username@domain
            pass: password

        username -- string -- the smtp username
        password -- string -- the smtp password
        domain -- string -- the domain the username belongs to
        """
        cli.run("echo \"{}\" | saslpasswd2 -c -u {} {} -p".format(password, domain, username))

        sasldb2 = Filepath("/etc/sasldb2")
        sasldb2.chmod(400)
        sasldb2.chown("postfix")




#!/usr/bin/env python
# http://docs.python.org/distutils/setupscript.html
# http://docs.python.org/2/distutils/examples.html

import sys
from setuptools import setup, find_packages
import ast

import stockton as module
name = module.__name__
version = module.__version__

setup(
    name=name,
    version=version,
    description='Easy peasy domain proxy',
    author='Jay Marcyes',
    author_email='jay@marcyes.com',
    url='http://github.com/jaymon/{}'.format(name),
    packages=find_packages(),
    #license="MIT",
    #install_requires=[''],
    tests_require=['testdata'],
    classifiers=[ # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Communications :: Email',
    ],
#     entry_points = {
#         'console_scripts': [
#             ' = {}'.format(name),
#         ],
#     },
)
